/* RSmtli4.sl - Wavefront MTL alike for illumination model 4 - FakeGlass
 * This shader is part of Ayam, a free 3D modeler for the RenderMan interface.
 * Copyright (c) 1998-2023 by Randolf Schultz (randolf.schultz@gmail.com)
 * All rights reserved. See License for details.
 * Parameters:
 * Ka - ambient reflectivity as color
 * Kd - diffuse reflectivity as color
 * Ks - specular reflectivity as color
 * Ns - specular exponent (1.0/roughness in RenderMan terms)
 * d  - dissolve, i.e. color independent opacity:
 *      1 - off, opaque
 *      0 - transparent
 *      - - transparent (with '-halo' option, creates opacity effect that is
 *          "dependent on the surface orientation relative to the viewer")
 * bm - bump multiplier
 * Ka_map   - multiplies Ka with colors from a texture file
 * Kd_map   - multiplies Kd with colors from a texture file
 * Ks_map   - multiplies Ks with colors from a texture file
 * Ns_map   - multiplies Ns with float from a texture file
 * d_map    - multiplies d with float from a texture file
 * bump_map - distort normal with value from a texture file
 * refl_map - reflection map
 */
surface
RSmtli4(color Ka = 1, Kd = .5, Ks = .5;
	float Ns = 1, d = 1, bm = 1;
	string Ka_map = "", Kd_map = "", Ks_map = "", Ns_map = "", d_map = "",
	       bump_map = "", refl_map = "";)
{
  normal Nf;
  vector V;
  color Kam = Ka, Kdm = Kd, Ksm = Ks;
  float Nsm = Ns, dm = abs(d);

  if(Ka_map != "")
    Kam *= color texture(Ka_map);

  if(Kd_map != "")
    Kdm *= color texture(Kd_map);

  if(Ks_map != "")
    Ksm *= color texture(Ks_map);

  if(Ns_map != "")
    Nsm *= float texture(Ns_map);

  if(d_map != "")
    dm *= float texture(d_map);

  if(bump_map != "")
    N = calculatenormal(P + bm * (float texture(bump_map)) * normalize(N));

  Nf = faceforward(normalize(N), I);
  V = -normalize(I);

  // adjust d according to specular highlight
  color Cspec = specular(Nf, V, 1.0/Nsm);
  float csmean = (comp(Cspec,0)+comp(Cspec,1)+comp(Cspec,2))/3.0;
  float ksmean = (comp(Ks,0)+comp(Ks,1)+comp(Ks,2))/3.0;

  if((d < 0) || (d == -0))
    {
      float sd = mix(dm, 1.0, min(acos(normalize(Nf).V)/1.5708,1));
      Oi = mix(sd, 1.0, max(csmean, ksmean));
    }
  else
    {
      Oi = mix(dm, 1.0, max(csmean, ksmean));
    }

  color Cr = 0.0;
  vector R = reflect(I, Nf);

  if(refl_map != "")
    Cr = environment(refl_map, vtransform("world", R));
  else
    Cr = trace(P, normalize(R));

  Ci = Oi * (Kam*ambient() + Kdm*diffuse(Nf) +
	     Ksm*(Cspec + Cr));

}
